/* fontsel.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <X11/Shell.h>	/* ?????? */

#include <X11/Xaw/Form.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>

#include "../common/fonts.h"
#include "../xcommon/dialogs.h"

enum {NO_GRP=0, FONT_GRP, SIZE_GRP, VADJ_GRP, HADJ_GRP};

static Widget topLevel, font_pshell, font_form,
	font_namelabel, font_namebox,
	font_attrlabel,
	font_attrbox, font_adjbox, font_adjlabel, font_borderbox,
	font_borderlabel, font_nobox, font_nolabel,
	font_sizelabel, font_sizebox,
	font_okbutton, font_cancelbutton;

typedef enum {nobox, namebox, borderbox, sizebox, attrbox, adjbox} boxtype;

static struct {
	Widget w;
	boxtype pw;
	char *text;
	int state;
	int grp;
	int value;
} buttons[] = {
	{0, namebox, "Courier", 1, FONT_GRP, COURIER},
	{0, namebox, "Helvetica", 0, FONT_GRP, HELVETICA},
	{0, namebox, "New Century Sch", 1, FONT_GRP, NEW_CENTURY},
	{0, namebox, "Times", 1, FONT_GRP, TIMES},
	{0, borderbox, "Top", 1, NO_GRP, BORDER_TOP},
	{0, borderbox, "Bottom", 1, NO_GRP, BORDER_BOTTOM},
	{0, borderbox, "Left", 1, NO_GRP, BORDER_LEFT},
	{0, borderbox, "Right", 1, NO_GRP, BORDER_RIGHT},
	{0, sizebox, "8", 1, SIZE_GRP, SIZE_8},
	{0, sizebox, "10", 0, SIZE_GRP, SIZE_10},
	{0, sizebox, "12", 1, SIZE_GRP, SIZE_12},
	{0, sizebox, "14", 1, SIZE_GRP, SIZE_14},
	{0, sizebox, "18", 1, SIZE_GRP, SIZE_18},
	{0, sizebox, "24", 1, SIZE_GRP, SIZE_24},
	{0, sizebox, "20", 1, SIZE_GRP, SIZE_20},
	{0, sizebox, "30", 1, SIZE_GRP, SIZE_30},
	{0, adjbox, "Top", 1, VADJ_GRP, VADJ_TOP},
	{0, adjbox, "Center", 0, VADJ_GRP, VADJ_CENTER},
	{0, adjbox, "Bottom", 1, VADJ_GRP, VADJ_BOTTOM},
	{0, adjbox, "Left", 0, HADJ_GRP, HADJ_LEFT},
	{0, adjbox, "Center", 1, HADJ_GRP, HADJ_CENTER},
	{0, adjbox, "Right", 1, HADJ_GRP, HADJ_RIGHT},
	{0, attrbox, "Bold", 1, NO_GRP, BOLD},
	{0, attrbox, "Italic", 1, NO_GRP, ITALIC}
};

#define NO_OF_BTNS (sizeof(buttons)/sizeof(buttons[0]))

enum {ABORT=0, DONE, WAITING};

static int status;

static void font_done(Widget w, XtPointer client_data, XtPointer call_data)
{
	XtPopdown(font_pshell);
	status = DONE;
}

static void font_abort(Widget w, XtPointer client_data, XtPointer call_data)
{
	XtPopdown(font_pshell);
	status = ABORT;
}

static void font_clicked(Widget w, XtPointer client_data, XtPointer call_data)
{
	int i, j;

	for (i = 0; i < NO_OF_BTNS; i++)
		if (w == buttons[i].w) break;
	if (i < NO_OF_BTNS) {
		if (buttons[i].grp != NO_GRP) {	/* is it a radio button? */
			for (j = 0; j < NO_OF_BTNS; j++) {
				if (j != i &&
					buttons[i].grp == buttons[j].grp &&
					buttons[j].state == 0) {
						buttons[j].state = 1;
						XtVaSetValues(buttons[j].w,
							XtNstate, !buttons[j].state,
							NULL);
					}
				buttons[i].state = 0;
			}
		} else {
			buttons[i].state = !buttons[i].state;
		}
		XtVaSetValues(w,
			XtNstate, !buttons[i].state, NULL);
	}
}

/* Returns the widget above the one we're creating */
static int from_vert(int i)
{
	int j;

	for (j = i-1; j >= 0; j--)
		if (buttons[j].pw == buttons[i].pw)
			return j;
	return -1;	/* happens for the first widget in each group */
}

static Widget parentbox(int i)
{
	switch (buttons[i].pw) {
	case nobox:
		return font_nobox;
	case namebox:
		return font_namebox;
	case sizebox:
		return font_sizebox;
	case attrbox:
		return font_attrbox;
	case borderbox:
		return font_borderbox;
	case adjbox:
		return font_adjbox;
	default:
		return font_adjbox;
	}
}

static Widget parentlabel(int i)
{
	switch (buttons[i].pw) {
	case nobox:
		return font_nolabel;
	case namebox:
		return font_namelabel;
	case sizebox:
		return font_sizelabel;
	case attrbox:
		return font_attrlabel;
	case borderbox:
		return font_borderlabel;
	case adjbox:
		return font_adjlabel;
	default:
		return font_adjlabel;
	}
}

void font_init(Widget tl)
{
	int i;

	topLevel = tl;
	font_pshell = XtVaCreatePopupShell("font_pshell",
		transientShellWidgetClass, topLevel, NULL);

	font_form = XtCreateManagedWidget("font_form",
		formWidgetClass, font_pshell, NULL, 0);

	font_namebox = XtVaCreateManagedWidget("font_namebox",
		formWidgetClass, font_form, NULL);
	font_namelabel = XtVaCreateManagedWidget("font_namelabel",
		labelWidgetClass, font_namebox, NULL);
	font_sizebox = XtVaCreateManagedWidget("font_sizebox",
		formWidgetClass, font_form,
		XtNfromHoriz, font_namebox, NULL);
	font_sizelabel = XtVaCreateManagedWidget("font_sizelabel",
		labelWidgetClass, font_sizebox, NULL);
	font_attrbox = XtVaCreateManagedWidget("font_attrbox",
		formWidgetClass, font_form,
		XtNfromHoriz, font_sizebox, NULL);
	font_attrlabel = XtVaCreateManagedWidget("font_attrlabel",
		labelWidgetClass, font_attrbox, NULL);
	font_borderbox = XtVaCreateManagedWidget("font_borderbox",
		formWidgetClass, font_form,
		XtNfromHoriz, font_attrbox, NULL);
	font_borderlabel = XtVaCreateManagedWidget("font_borderlabel",
		labelWidgetClass, font_borderbox, NULL);
	font_adjbox = XtVaCreateManagedWidget("font_adjbox",
		formWidgetClass, font_form,
		XtNfromHoriz, font_borderbox, NULL);
	font_adjlabel = XtVaCreateManagedWidget("font_adjlabel",
		labelWidgetClass, font_adjbox, NULL);
	for (i = 0; i < NO_OF_BTNS; i++) {
		char b[256];
		int vert;

		sprintf(b, "font_button%d", i);
		buttons[i].w = XtVaCreateManagedWidget(b,
			toggleWidgetClass, parentbox(i),
			XtNradioData, buttons[i].value,
			XtNtop, XtChainTop,
			XtNbottom, XtChainTop,
			XtNleft, XtChainLeft,
			XtNright, XtChainRight,
			XtNlabel, buttons[i].text,
			NULL);
		if ((vert = from_vert(i)) >= 0) {
			XtVaSetValues(buttons[i].w,
				XtNfromVert, buttons[vert].w, NULL);
		} else {
			XtVaSetValues(buttons[i].w,
				XtNfromVert, parentlabel(i), NULL);
		}
		XtAddCallback(buttons[i].w, 
			XtNcallback, font_clicked, (XtPointer)i);
	}

	font_okbutton = XtVaCreateManagedWidget("font_okbutton",
		commandWidgetClass, font_form, 
		XtNtop, XtChainTop,
		XtNbottom, XtChainTop,
		XtNleft, XtChainRight,
		XtNright, XtChainRight,
		XtNfromHoriz, font_adjbox, NULL);
	XtAddCallback(font_okbutton, XtNcallback, font_done, NULL);
	font_cancelbutton = XtVaCreateManagedWidget("font_cancelbutton",
		commandWidgetClass, font_form, 
		XtNtop, XtChainTop,
		XtNbottom, XtChainTop,
		XtNleft, XtChainRight,
		XtNright, XtChainRight,
		XtNfromHoriz, font_adjbox,
		XtNfromVert, font_okbutton, NULL);
	XtAddCallback(font_cancelbutton, XtNcallback, font_abort, NULL);
}

static void set_button(int group, int value)
{
	int i;

	for (i = 0; i < NO_OF_BTNS; i++)
		if (buttons[i].grp == group && buttons[i].value == value)
			buttons[i].state = 0;
}

int font_input(int *format, int *mask)
{
	int i;

	*mask = 0;	/* no fields have been changed yet */
	/* set buttons to what's already there */
	for (i = 0; i < NO_OF_BTNS; i++)
		buttons[i].state = 1;
	set_button(FONT_GRP, *format & FONT_MASK);
	set_button(SIZE_GRP, *format & SIZE_MASK);
	set_button(NO_GRP, *format & ITALIC);
	set_button(NO_GRP, *format & BOLD);
	set_button(NO_GRP, *format & BORDER_LEFT);
	set_button(NO_GRP, *format & BORDER_RIGHT);
	set_button(NO_GRP, *format & BORDER_TOP);
	set_button(NO_GRP, *format & BORDER_BOTTOM);
	set_button(HADJ_GRP, *format & HADJ_MASK);
	set_button(VADJ_GRP, *format & VADJ_MASK);
	set_button(NO_GRP, *format & INVERT);
	for (i = 0; i < NO_OF_BTNS; i++)
		XtVaSetValues(buttons[i].w,
			XtNstate, !buttons[i].state, NULL);
	status = WAITING;

	center(topLevel, font_pshell);
	XtPopup(font_pshell, XtGrabNonexclusive);
	while (status == WAITING) {
		XtAppContext app_context = XtWidgetToApplicationContext(topLevel);
		XEvent event_return;
		XtAppNextEvent(app_context, &event_return);
		XtDispatchEvent(&event_return);
	}
	*format = 0;
	for (i = 0; i < NO_OF_BTNS; i++)
		if (!buttons[i].state) *format |= buttons[i].value;
	return status;
}

